; ==================================================================================================
; Title:   COM_Dispatch.asm
; Author:  G.Friedrich
; Version: 1.0.0
; Purpose: COM dispatching helper procedures compatible with ObjAsm32 definitions.
; Notes:   Version 1.0.0, January 2006
;           - First release. Source code hints taken from Japheth's DispHelp.inc.
; ==================================================================================================


LOCALE_SYSTEM_DEFAULT equ 800h

.code

; 
; Procedure: ExecDispatch
; Purpose:   Performs a method call using the IDispatch interface. Variant arguments are
;            automatically created onto stack.
; Arguments: Arg 1: -> Dispatch interface.
;            Arg 2: Method Dispatch ID.
;            Arg 3: Dispatch type (DISPATCH_METHOD, DISPATCH_PROPERTYGET, DISPATCH_PROPERTYPUT,
;                    DISPATCH_PROPERTYPUTREF).
;            Arg 4: Argument count.
;            Arg 5: Return Argument size in bytes.
;            Arg 6 - n: Argument types.
;            Arg n+1 - m: Argument values.
;            Arg m+1: -> Return Value.

ExecDispatch proc C uses edi esi, pDispatch:Pointer, dDispID:dword, dDispType:dword,
                                  dArgCount:dword, dRetType:dword, Args:vararg
    local dESP:dword, dArgIndexError:dword
    local dPropPutID:dword, DispParams:DISPPARAMS

    mov dESP, esp

    ;Setup the VARIANT argument array. The varints have to be arranged in REVERSE order!
    mov ecx, dArgCount
    mov DispParams.cArgs, ecx
    lea esi, Args                                   ;esi -> argument types
    lea edx, [esi + ecx * 4]                        ;All passed VT_xxx are 4 byte sized
    .while ecx != 0
      sub esp, sizeof VARIANT
      lea edi, [esp + 8]                            ;edi -> Variant data
      lodsd
      mov [esp].VARIANT.vt, ax
      xchg edx, esi
      .if (eax & VT_BYREF)
        movsd
      .elseif (eax == VT_VARIANT)
        sub edi, 8
        movsd
        movsd
        movsd
        movsd
      .elseif (eax == VT_R8 || eax == VT_DATE || eax == VT_CY)
        movsd
        movsd
      .else
        movsd
      .endif
      xchg esi, edx
      dec ecx
    .endw
    mov edi, edx                                    ;edi -> Return value
    mov DispParams.rgvarg, esp

    ;Setup the VARIANT return value
    .if dRetType != -1
      sub esp, sizeof VARIANT
      mov esi, esp                                  ;esi -> return Variant
      mov [esp].VARIANT.vt, VT_EMPTY
    .else
      xor esi, esi
    .endif

    ;Setup the DISPPARAMS structure
    .if ((dDispType == DISPATCH_PROPERTYPUT) || (dDispType == DISPATCH_PROPERTYPUTREF))
      mov dPropPutID, DISPID_PROPERTYPUT
      lea eax, dPropPutID
      mov DispParams.rgdispidNamedArgs, eax
      mov DispParams.cNamedArgs, 1
    .else
      mov DispParams.cNamedArgs, 0
    .endif

    ;Call the Dispatch.Invoke method
    ICall pDispatch::IDispatch.DispatchInvoke, dDispID, addr IID_NULL, LOCALE_SYSTEM_DEFAULT, \
            dDispType, addr DispParams, esi, NULL, addr dArgIndexError

    ;Copy the value of the returned Variant to the return Value
    .if (eax == S_OK && esi)
      mov edi, [edi]                                ;Get return argument pointer
      .if (dRetType == VT_VARIANT)
        movsd
        movsd
        movsd
        movsd
      .else
        add esi, 8
        .if (dRetType == VT_R8 || dRetType == VT_DATE || dRetType == VT_CY)
          movsd
          movsd
        .else
          movsd
        .endif
      .endif
    .endif

    mov esp, dESP                                   ;Restore stack
    mov ecx, dArgIndexError                         ;eax = ICall result value, ecx = index error
    ret
ExecDispatch endp

; 
; Procedure: CallDispatch
; Purpose:   Performs a method call using the IDispatch interface. Variant arguments are passed as
;            procedure arguments. The caller has to assembe the variants.
; Arguments: Arg 1: -> Dispatch interface.
;            Ar2 2: Method Dispatch ID.
;            Arg 3: Dispatch type (DISPATCH_METHOD, DISPATCH_PROPERTYGET, DISPATCH_PROPERTYPUT,
;                     DISPATCH_PROPERTYPUTREF).
;            Arg 4: Number of arguments.
;            Arg 5: -> variant arguments.
;            Arg 6: -> Return Variant.

CallDispatch proc pDispatch:Pointer, dDispID:dword, dDispType:dword, dArgNum:dword, \
                  pVarArgs:Pointer, pVarRet:Pointer
    local dPropPutID:dword, DispParams:DISPPARAMS, dArgIndexError:dword

    ;Setup the DISPPARAMS structure
    m2m DispParams.cArgs, dArgNum, eax
    m2m DispParams.rgvarg, pVarArgs, eax

    .if ((dDispType == DISPATCH_PROPERTYPUT) || (dDispType == DISPATCH_PROPERTYPUTREF))
      mov dPropPutID, DISPID_PROPERTYPUT
      lea eax, dPropPutID
      mov DispParams.rgdispidNamedArgs, eax
      mov DispParams.cNamedArgs, 1
    .else
      mov DispParams.cNamedArgs, 0
    .endif

    ;Call the Dispatch.Invoke method    LOCALE_SYSTEM_DEFAULT = 800h
    ICall pDispatch::IDispatch.DispatchInvoke, dDispID, addr IID_NULL, LOCALE_SYSTEM_DEFAULT, \
            dDispType, addr DispParams, pVarRet, NULL, addr dArgIndexError
    mov ecx, dArgIndexError                         ;eax = ICall result value, ecx = index error
    ret
CallDispatch endp
